﻿"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.generatePlugin = void 0;
const tslib_1 = require("tslib");
const common_1 = require("@kitajs/common");
const json_stable_stringify_1 = tslib_1.__importDefault(require("json-stable-stringify"));
const ts_writer_1 = require("ts-writer");
const path_1 = require("../util/path");
function generatePlugin(routes, schemas, plugins, providers, cwdSrcRelativity) {
    return (0, ts_writer_1.ts) `${'plugin'}
    'use strict';

    const fp = require('fastify-plugin');

    /**
     * The Kita generated fastify plugin. Registering it into your fastify instance will
     * automatically register all routes, schemas and providers.
     *
     * @example
     * \`\`\`ts
     * import { Kita } from '@kitajs/runtime';
     * 
     * const app = fastify();
     * 
     * app.register(Kita)
     * 
     * app.listen().then(console.log);
     * \`\`\`
     * 
     * @see {@link https://kita.js.org}
     */
    exports.Kita = fp(
      async (${common_1.kFastifyVariable}, ${common_1.kKitaOptions}) => {
        // Register all plugins
${plugins.map(toPlugin)}

        // Import all routes
${routes.map(toRoute)}

        // Import all providers with application hooks
${providers.filter((p) => p.applicationHooks.length).map((p) => toProvider(p, cwdSrcRelativity))}

        // Add all schemas
${schemas.map(toSchema)}

        // Register all routes - inside a plugin to make sure capsulation works
        // https://github.com/fastify/fastify-plugin/issues/78#issuecomment-672692334
        await ${common_1.kFastifyVariable}.register(async ${common_1.kFastifyVariable} => {
${providers.flatMap((p) => p.applicationHooks.map((h) => `${common_1.kFastifyVariable}.addHook('${h}', ${p.type}.${h});`))}

${routes.map((r) => `${common_1.kFastifyVariable}.route(${r.schema.operationId}.${r.schema.operationId}Options)`)}
        }, ${common_1.kKitaOptions});
      },
      {
        name: 'Kita',
        fastify: '4.x'
      }
    );

    ${getSchemaDefinitions(schemas) && `module.exports.${common_1.kSchemaDefinitions} = ${toSchemaDefinitions(schemas)}`}

    exports.__esModule = true;

    ${ts_writer_1.ts.types}

    import type { FastifyPluginAsync } from 'fastify'

    export declare const Kita: FastifyPluginAsync<{${plugins.map(toPluginType).join(',\n')}}>;
    
    ${getSchemaDefinitions(schemas) &&
        `
/**
 * All schemas exported in the \`providers/schemas.ts\` file
 */
export declare const ${common_1.kSchemaDefinitions}: ${toSchemaDefinitions(schemas)}
   `}
  `;
}
exports.generatePlugin = generatePlugin;
function toRoute(route) {
    return `const ${route.schema.operationId} = require('./routes/${route.schema.operationId}');`;
}
function toProvider(provider, cwdSrcRelativity) {
    return `const ${provider.type} = require(${(0, path_1.toMaybeRelativeImport)(provider.providerPath, cwdSrcRelativity)});`;
}
function getSchemaDefinitions(schemas) {
    return schemas.find((schema) => schema.$id === common_1.kSchemaDefinitions);
}
function toSchemaDefinitions(schemas) {
    const kitaSchema = getSchemaDefinitions(schemas);
    if (!kitaSchema) {
        return '{}';
    }
    return (0, json_stable_stringify_1.default)({
        ...kitaSchema,
        // Remove the $id to make it more readable without messing up the schema
        $id: undefined
    }, { space: 4 });
}
function toPlugin(plugin) {
    return `
   if (${common_1.kKitaOptions}.${plugin.name} !== false) {
     await ${common_1.kFastifyVariable}.register(require("${plugin.importUrl}"), Object.assign(${(0, common_1.stringifyOptions)(plugin.options)}, ${common_1.kKitaOptions}.${plugin.name} || {}))
   }
  `;
}
function toSchema(schema) {
    return `${common_1.kFastifyVariable}.addSchema(${(0, json_stable_stringify_1.default)(schema, { space: 4 })})`;
}
function toPluginType(plugin) {
    return `
    /**
     * Options for the \`${plugin.name}\` plugin. Use \`false\` to disable it manually.
     * 
     * @see {@link ${plugin.importUrl}}
     */
    ${plugin.name}?: Parameters<typeof import("${plugin.importUrl}").default>[1] | false`;
}
//# sourceMappingURL=plugin.js.map